import threading
import queue
import time

class ThreadedSerialReader:
    """A class that monitors a serial port using a separate thread.
    It returns the received lines to a given callback.
    It can be paused and resumed asynchronously.
    
    This class was inspired by the threaded reader and process serial stream from the official pyserial docs.
    see: https://www.pyserial.org/docs/reading-data
    
    The addition that the thread can be paused and resumed asynchronously was not originally in there.
    """

    def __init__(self, ser,  received_line_handler=lambda received: print(f'Received: {received}')):
        self.ser = ser
        self.data_queue = queue.Queue(maxsize=1000)
        self.running = True
        self.thread = None
        self.blocked = threading.Event() # Own addition to allow for pausing of the ThreadedSerialReader.
        self.received_line_handler = received_line_handler

    def pause(self):
        self.blocked.set()
    
    def resume(self):
        self.blocked.clear()
    
    def is_paused(self):
        return self.blocked.is_set()
    
    def start(self):
        """Start background reading thread"""
        self.thread = threading.Thread(target=self._read_loop)
        self.thread.daemon = True
        self.thread.start()
    
    def _read_loop(self):
        """Background reading loop"""
        line_buffer = b''

        while self.running:
            try:
                if not self.blocked.is_set() and self.ser.in_waiting:
                    chunk = self.ser.read(self.ser.in_waiting)
                    line_buffer += chunk
                    
                    # Process complete lines
                    while b'\n' in line_buffer:
                        line, line_buffer = line_buffer.split(b'\n', 1)
                        try:
                            text = line.decode('utf-8').strip()
                            if text:
                                # Parse the message line
                                self.received_line_handler(text)
                        except Exception as e:
                            print(f"Process error: {e}")
                else:
                    time.sleep(0.001)  # Small delay when no data
            except Exception as e:
                if self.running:
                    print(f"Read thread error: {e}")
    
    def get_data(self, timeout=0.1):
        """Get data from queue"""
        try:
            return self.data_queue.get(timeout=timeout)
        except queue.Empty:
            return None
    
    def get_all_data(self):
        """Get all queued data"""
        data = []
        while not self.data_queue.empty():
            try:
                data.append(self.data_queue.get_nowait())
            except queue.Empty:
                break
        return b''.join(data)

    def stop(self):
        """Stop reading thread"""
        self.running = False
        if self.thread:
            self.thread.join(timeout=None)
    
    def __del__(self):
        """Stop the running thread when the object gets deleted"""
        self.stop()

"""
if __name__ == '__main__':
    # EXAMPLE USAGE
    ser = serial.Serial(port="COM5", baudrate=9600, timeout=1)
    reader = ThreadedSerialReader(ser)
    reader.start()

    reader.pause()
    print(write_with_confirm(ser, b'AT\r\n', b'+AT: OK'))
    print(write_with_confirm(ser, b'AT+MODE=TEST\r\n', b'+MODE: TEST'))
    print(write_with_confirm(ser, b'AT+TEST=RXLRPKT\r\n', b'+TEST: RXLRPKT'))
    reader.resume()

    try:
        while True:
            data = reader.get_data(timeout=1)
            if data:
                print(f"Got {len(data)} bytes: {data}")
            else:
                print("No data received")
            time.sleep(0.1)
    finally:
        reader.stop()
"""


import time

from lora_node import LoRaNode

if __name__ == "__main__":
    node1 = LoRaNode("COM4")
    time.sleep(2)  # Give the nodes some time to set up and start listening
    node2 = LoRaNode("COM5")

    time.sleep(2)  # Give the nodes some time to set up and start listening
    node1.send("AABBCCDD")



import threading
import time
from queue import Queue

from .lora_node import LoRaNode
from .receiving.received_message_data_parser import ReceivedMessage
from .receiving.response import Response
from .transmission import Transmission, TransmissionState

class ReliableCommunicatingNode:
    """
    A reliable communicating node for sending messages over LoRa.
    Uses acknowledgements and retransmissions to ensure message delivery."""

    def __init__(
            self,
            lora_node: LoRaNode,
            incoming_message_handler = lambda _: b"OK!" # Accepts an incoming ReceivedMessage, and returns OK! as an answer
    ):
        """incoming messages to a callback that handles ReceivedMessages.
        
        Args:
            lora_node: An instantiated LoRaNode that this class uses to communicate. This is just a layer above the lora node.
            max_retries: The maximum number of times that a message is resend when no ACK is retured.
            incoming_message_handler: A callback that accepts instances of ReceivedMessage, and returns a reply to be sent back. This callback is called whenever a message arrives."""

        self.lora_node = lora_node
        self.incoming_message_handler = incoming_message_handler

        self.lora_node.set_on_received_callback(self._on_receive)
        
        self.send_queue : Queue = Queue() # Queue of transmissions to be sent reliably

        self.current_transmission : Transmission = None # The transmission that is currently being sent/waiting for ACK

    def send_reliably(self, data: bytes, max_retries: int = 3, retransmission_timeout: float = 5.0):
        """Sends a message using best effort, but non-blocking.
        If no response is sent within the retransmission timeout,
        it will retry until the number of max retries is reached."""

        # Create new transmission object and add it to the send queue
        transmission = Transmission(data, max_retries, retransmission_timeout)
        self.send_queue.put(transmission)

        # If there is no current transmission being sent, start handling the next one in the queue
        if self.current_transmission is None:
            self._handle_next_in_send_queue()
        else:
            print("Currently busy sending another message, adding to send queue")
    
    def send_reliably_wait_for_answer(self, data: bytes, max_retries: int = 3, retransmission_timeout: float = 5.0) -> bytes:
        """Sends a message using best effort and blocks until a response is received or the max retries are reached.
        If the max retries are reached without any response, a TimeoutException is raised."""

        # Create new transmission object and add it to the send queue
        transmission = Transmission(data, max_retries, retransmission_timeout)
        self.send_queue.put(transmission)

        # If there is no current transmission being sent, start handling the next one in the queue
        if self.current_transmission is None:
            self._handle_next_in_send_queue()
        else:
            print("Currently busy sending another message, adding to send queue")
        
        transmission.terminated.wait(timeout=None) # Wait until the transmission is finished successfully or unsuccessfully.

        if transmission.state == TransmissionState.FAILED:
            # First clear the transmission and send the next message in queue before throwing an error.
            # Else the queue would become blocked.
            self.current_transmission = None
            self._handle_next_in_send_queue()
            raise TimeoutError("Connection timed out: The number of max retransmissions was reached, without receiving a reply.")
        else: # The transmission was successful. Return the reply
            return transmission.get_response()
    
    def _handle_next_in_send_queue(self):
        if not self.send_queue.empty():
            next_transmission = self.send_queue.get()
            self.current_transmission = next_transmission
            self._transmit_current()
    
    def _transmit_current(self):
        transmission = self.current_transmission
        self.lora_node.send(transmission.send_data)
        # Start the retransmission timer in a separate thread so that we can wait for the ACK without blocking the main thread.
        # If the timer expires before an ACK is received, it will trigger a retransmission if the max retries has not been reached.
        timer_thread = threading.Thread(
            target=transmission.retransmission_timer,
            args=(self._transmit_current,))
        timer_thread.start()

    def _on_receive(self, message: ReceivedMessage):
        # Process the received message and send an acknowledgement back to the sender
        if message.has_payload():
            payload = message.get_payload()
            # Check if the received message is a response to an earlier sent message.
            payload_as_response = Response.from_bytes(payload)
            if payload_as_response is not None:
                # The payload could be interpreted as a response.
                # Check if the response is for the last sent message.
                if payload_as_response.is_response_for(self.current_transmission.send_data):
                    print(f"Received: {str(payload_as_response)}")
                    self.current_transmission.mark_acknowledged(payload_as_response.get_contents())
                    self.current_transmission = None # Clear the current transmission before handling the next one in the queue
                    self._handle_next_in_send_queue()
                else:
                    print(f"⚠️ WARNING: received response to an unknown message: {payload_as_response.get_original_message()}. Expected a response for {self.current_transmission.send_data}")
            else:
                # The received message is not a response to a message that this node sent.
                # Send an acknowledgement back to the sender for the received message
                answer: bytes = self.incoming_message_handler(message)
                # Send a response back to the sender
                resp = Response(response_for=message, response_contents=answer)
                # Just send, without expecting a reply to this reply
                self.lora_node.send(resp.as_bytes())
        else:
            print(f"⚠️ WARNING: received message without payload: {message}")


def testReliableCommunicatingNode():
    # Some end-to-end tests for ReliableCommunicatingNode
    
    node1 = LoRaNode(port="COM4")
    reliable_node1 = ReliableCommunicatingNode(node1)

    # The receiving node is not yet initialized, so this should time out.
    try:
        reliable_node1.send_reliably_wait_for_answer(b"Should not arrive", 1, 0.5)
        assert False, "No error was thrown, but the message couldn't have arrived."
    except TimeoutError:
        print("Success: the message that couldn't arrive, did not arrive.")

    # Initialize a second node and do some more tests.
    node2 = LoRaNode(port="COM5")
    reliable_node2 = ReliableCommunicatingNode(node2)

    answer = reliable_node2.send_reliably_wait_for_answer(b"LOL", 1, 0.5)
    print(f"WAITED FOR AND GOT {answer}")

    reliable_node1.send_reliably(b"HELLO WORLD")
    print("DID SOME WORK WHILE WAITING")
    time.sleep(2)


class ReceivedMessage:
    def __init__(self, metadata = {}, hexpayload = None):
        self.metadata = metadata # Dictionary to hold metadata key-value pairs
        # The actual payload of the message can be set later too.
        self.payload = None
        if hexpayload is not None:
            self.set_payload(hexpayload)

    def set_payload(self, hexpayload: str):
        self.payload = bytes.fromhex(hexpayload)
    
    def has_payload(self):
        return self.payload is not None
    
    def get_payload(self) -> bytes:
        return self.payload
    
    def __str__(self):
        return f"ReceivedMessage(metadata={self.metadata}, payload={self.payload})"



from .reliable_communicating_node import testReliableCommunicatingNode

# Run this file using `python -m lora_p2p`

if __name__ == "__main__":
    # INTEGRATION TEST
    testReliableCommunicatingNode()


import serial

from .lora_kit.lora_kit_controller import LoRaKitController
from .lora_kit.mock_lora_kit_controller import MockLoRaKitController
from .receiving.received_message_data_parser import ReceivedMessage

class LoRaNode:
    """A high level communicating node that uses a LoRaKitController to take care of switching modes and setting up the hardware.
    Can send messages to other nodes and also handles incoming ReceivedMessages from other nodes."""

    def __init__(self, port: None | str = None):
        """Creates an instance of a LoRaNode.
        
        Args:
            port: the name of the physical port that the LoRa Wio-E5 Development Kit is connected to.
                If this is None, a mock version of a LoRa node is created that doesn't use real hardware."""

        if port is None:
            # Create a mocked version of the LoRaKitController. This works without real hardware.
            self.lora_controller = MockLoRaKitController(received_message_handler=self.receive)
        else:
            # Use real hardware connected at the specified port.
            ser = serial.Serial(port, baudrate=9600, timeout=1)
            self.lora_controller = LoRaKitController(ser, received_message_handler=self.receive)
        
        # Setting up the LoRa module
        # Test connection and set to test mode
        self.lora_controller.check_connection()
        self.lora_controller.enable_test_mode()
        self.lora_controller.enable_listening()

    def set_on_received_callback(self, callback) -> None:
        self.on_received_callback = callback

    def send(self, data: bytes):
        """Sends a message, which disables listening.
        Resumes listening after finishing sending"""

        self.lora_controller.send_message(data) # Send the message, which disables listening.
        self.lora_controller.enable_listening() # Resume listening after sending

    def receive(self, message: ReceivedMessage):
        """Handles receiving a message. This should only be called when the node is in listening mode."""
        # Implement receiving logic here
        if not self.is_listening():
            print("Cannot receive while not listening.")
            return
        if self.on_received_callback:
            self.on_received_callback(message)
        else:
            print("Received a message, but no callback is set.")

    def is_listening(self):
        return self.lora_controller.is_listening()




import serial
import time
from serial_write_with_confirm import write_with_confirm
from threaded_serial_reader import ThreadedSerialReader

if __name__ == '__main__':
    ser1 = serial.Serial(port="COM5", baudrate=9600, timeout=1)
    reader = ThreadedSerialReader(ser1)
    reader.start()

    reader.pause()
    print(write_with_confirm(ser1, b'AT\r\n', b'+AT: OK'))
    print(write_with_confirm(ser1, b'AT+MODE=TEST\r\n', b'+MODE: TEST'))
    print(write_with_confirm(ser1, b'AT+TEST=RXLRPKT\r\n', b'+TEST: RXLRPKT'))
    reader.resume()

    ser2 = serial.Serial(port="COM4", baudrate=9600, timeout=1)
    print(write_with_confirm(ser2, b'AT\r\n', b'+AT: OK'))
    print(write_with_confirm(ser2, b'AT+MODE=TEST\r\n', b'+MODE: TEST'))

    # MULTI LINE EXPECTED RESPONSE TEST
    message = "123456789ABCDEF"
    # Add a 0 in the beginning if the hex message length is uneven
    if len(message)%2 == 1:
        message = "0" + message
    
    command = (f'AT+TEST=TXLRPKT, "{message}"\r\n').encode()
    response = (f'+TEST: TXLRPKT "{message}"\r\n+TEST: TX DONE').encode()
    
    # Send the message from ser2
    print(write_with_confirm(ser2, command, response))

    # Give the response some time to arrive
    time.sleep(3)
    """
    # Read the response from ser1
    try:
        while True:
            data = reader.get_data(timeout=1)
            if data:
                print(f"Got {len(data)} bytes: {data}")
            else:
                print("No data received")
            time.sleep(0.1)
    finally:
        reader.stop()
    """




    



import time

import serial
import string
from .serial_helper_code.threaded_serial_reader import ThreadedSerialReader
from .serial_helper_code.serial_write_with_confirm import write_with_confirm

from ..receiving.received_message_data_parser import ReceivedMessageDataParser


class LoRaKitController:
    """Class that provides a high level interface for sending AT commands to a LoRa module
    over a serial connection, using pyserial."""

    def __init__(self, ser: serial.Serial, received_message_handler=lambda x: print(x)):

        self.ser = ser
        self.received_message_handler = received_message_handler
        
        # Set the callback that handles the received messages
        self.received_message_data_handler = ReceivedMessageDataParser()

        self.threaded_serial_reader = ThreadedSerialReader(
            ser,
            received_line_handler=self.handle_incoming_message_line
        )
        self.threaded_serial_reader.start()

    def check_connection(self) -> bool:
        return self._write_command_and_check_response(b'AT\r\n', b'+AT: OK')
    
    def enable_test_mode(self) -> bool:
        return self._write_command_and_check_response(b'AT+MODE=TEST\r\n', b'+MODE: TEST')
    
    def enable_listening(self) -> bool:
        success = self._write_command_and_check_response(b'AT+TEST=RXLRPKT\r\n', b'+TEST: RXLRPKT')
        if not success:
            raise RuntimeError("Couldn't enter listening mode")
        else:
            # IMPORTANT: resume the listening thread
            self.threaded_serial_reader.resume()
            return success
    
    def send_message(self, payload: bytes):
        # Need to send the payload as a hex string via the AT commands.
        hex_message = payload.hex()
        # Check if the message is a hex string.
        assert all(c in string.hexdigits for c in hex_message)
        # Add a 0 in the beginning if the hex message length is uneven
        if len(hex_message)%2 == 1:
            hex_message = "0" + hex_message
        
        # letters to upper case, because the received response always is received in upper case
        hex_message = hex_message.upper()
        
        command = (f'AT+TEST=TXLRPKT, "{hex_message}"\r\n').encode()
        response = (f'+TEST: TXLRPKT "{hex_message}"\r\n+TEST: TX DONE').encode()
    
        return self._write_command_and_check_response(command, response)

    def handle_incoming_message_line(self, line):
        result = self.received_message_data_handler.process_message_line(line)
        if result:
            # Give the message to the callback that handles the ReceivedMessages.
            self.received_message_handler(result)

    def _write_command_and_check_response(self, command, expected_response) -> bool:
        ####if not self.threaded_serial_reader.is_paused():
        ####    raise RuntimeError("should first pause the threaded serial reader before trying to write a command.")
        self.threaded_serial_reader.pause()

        (success, _, response) = write_with_confirm(self.ser, command, expected_response)

        if not success:
            print(f"⚠️ WARNING: AT command '{command}' got unexpected response: '{response}'. Expected '{expected_response}' instead.")
        return success
    
    def is_listening(self) -> bool:
        return not self.threaded_serial_reader.is_paused()


if __name__ == '__main__':
    ser = serial.Serial('COM4')
    helper = LoRaKitController(ser)

    print(f"Connection OK: {helper.check_connection()}")
    print(f"Test mode OK: {helper.enable_test_mode()}")
    print(f"Listening OK: {helper.enable_listening()}")
    time.sleep(1)


    ser2 = serial.Serial('COM5')
    helper2 = LoRaKitController(ser2)

    print(f"Connection OK: {helper2.check_connection()}")
    print(f"Test mode OK: {helper2.enable_test_mode()}")
    print(f"Sending OK: {helper2.send_message(bytes.fromhex("0123456789ABCDEF"))}")
    time.sleep(1)

    # TODO Check if the message was correctly received


    





import time
import serial

def write_with_confirm(ser, data, expected_response=b'OK', timeout=2):
    """Write data and wait for confirmation see https://www.pyserial.org/docs/writing-data"""
    # Clear input buffer
    ser.reset_input_buffer()
    
    # Send data
    bytes_written = ser.write(data)
    ser.flush()  # Force transmission
    
    # Wait for confirmation
    start_time = time.time()
    response = b''
    
    while time.time() - start_time < timeout:
        if ser.in_waiting:
            chunk = ser.read(ser.in_waiting)
            response += chunk
            
            if expected_response in response:
                return True, bytes_written, response
        
        time.sleep(0.01)
    
    return False, bytes_written, response

if __name__ == '__main__':
    # TEST EXAMPLE USAGE
    ser = serial.Serial(port="COM4", baudrate=9600, timeout=1)
    print(write_with_confirm(ser, b'AT\r\n', b'+AT: OK'))
    print(write_with_confirm(ser, b'AT+MODE=TEST\r\n', b'+MODE: TEST'))

    # MULTI LINE EXPECTED RESPONSE TEST
    message = "123456789ABCDEF"
    # Add a 0 in the beginning if the hex message length is uneven
    if len(message)%2 == 1:
        message = "0" + message
    
    command = (f'AT+TEST=TXLRPKT, "{message}"\r\n').encode()
    response = (f'+TEST: TXLRPKT "{message}"\r\n+TEST: TX DONE').encode()

    print(write_with_confirm(ser, command, response))

    
    


import threading
from enum import Enum

class TransmissionState(Enum):
    UNACKNOWLEDGED = 1
    ACKNOWLEDGED = 2
    FAILED = 3

class Transmission():
    """
    Represents a message transmission with its state and retry information.
    Has a threading event `terminated` which will be signaled when the transmission was successful or reached max retries.
    """

    def __init__(self, send_data: bytes, max_retries: int, timeout: float):
        assert isinstance(send_data, bytes), "the data to be sent should be bytes"
        assert isinstance(max_retries, int) and max_retries >= 0, "max retries should be an integer >= 0"
        assert isinstance(timeout, float) and timeout > 0, "retransmission timeout should be a float > 0"

        self.send_data = send_data
        self.max_retries = max_retries
        self.timeout = timeout

        self.state = TransmissionState.UNACKNOWLEDGED
        self.response_payload = None

        self.retries = 0
        self.terminated = threading.Event() # Event that signals that the transmission is finished. Either response received or reached max retries.

    def mark_acknowledged(self, response_payload: bytes):
        self.state = TransmissionState.ACKNOWLEDGED
        self.response_payload = response_payload
        self.terminated.set()
    
    def get_response(self):
        assert self.response_payload is not None, "tried to access a response to a transmission that hasn't arrived (yet)"
        return self.response_payload

    def _mark_unsuccessful(self):
        self.state = TransmissionState.FAILED
        self.terminated.set()
    
    def retransmission_timer(self, retransmit_callback):
        """
        Starts a timer to wait for an acknowledgement. If the timer expires before an ACK is received, it will trigger a retransmission if the max retries has not been reached."""        
        
        print("Timer started")

        # Wait until ACK received or timeout
        if self.terminated.wait(self.timeout):
            print("ACK received -> cancel retransmission")
            return

        # Timeout occurred. mark as failed if we have reached the max retries.
        self.retries += 1
        if (self.retries > self.max_retries):
            print("Max retries reached -> marking transmission as failed")
            self._mark_unsuccessful()
        else:
            print("Timeout -> retransmitting")
            retransmit_callback()


import threading
from typing import override
import time

from .lora_kit_controller import LoRaKitController
from ..receiving.received_message_data_parser import ReceivedMessage

class MockMedium:
    """A fake medium over wich MockSerialHelpers can communicate."""

    def __init__(self):
        # TODO: add options to add artificial delay and other parameters.
        self.lora_controllers: list[MockLoRaKitController] = []

    def join(self, lora_controller):
        """Function to let a MockLoRaKitController join the medium.
        This allows it to receive messages from other MockLoRaKitController."""

        self.lora_controllers.append(lora_controller)
    
    def broadcast(self, source, payload: bytes):
        """Broadcast a message payload from a MockLoRaKitController to the others on the medium."""
        # The received hex payload for the real LoRa modules is always in upper case, and if uneven, starts with a zero.
        hex_payload = payload.hex().upper()
        packet_length = len(hex_payload)
        if len(hex_payload) % 2 == 1:
            hex_payload = "0" + hex_payload
            packet_length += 1
        packet_length /= 2 # Packet length is always the hex string length / 2

        # Construct a received message.
        message_to_be_received = ReceivedMessage(
            metadata={"LEN": packet_length, "RSSI": None, "SNR": None}, # TODO mock these values too in a realistic way.
            hexpayload=hex_payload
        )

        for lora_controller in self.lora_controllers:
            if lora_controller == source:
                continue # Don't send to the source

            # Deliver the message using a separate thread.
            messenger_thread = threading.Thread(
                target=self.deliver_message,
                args=(lora_controller, message_to_be_received,)
            )
            messenger_thread.start()
    
    def deliver_message(self, lora_controller, message_to_be_received: ReceivedMessage):
        """Delivers a message to a receiving callback of an LoRaKitController after a small delay.
        Is supposed to be executed in a separate thread."""
        # Add a small transmission delay
        time.sleep(0.05)
        if lora_controller.is_listening():
            lora_controller.received_message_handler(message_to_be_received)
        else:
            raise RuntimeError("The receiving mocked lora controller was not listening while another node tried to deliver a message")


class MockLoRaKitController(LoRaKitController):
    """Class that mocks the behaviour of the LoRaKitController class which controls a LoRa module.
    This can be used for testing purposes without needing to have an actual LoRa module connected via a serial connection."""

    # Medium as a static variable, shared by all LoRaKitController instances.
    medium = MockMedium()

    def __init__(self, received_message_handler=lambda x: print(x)):
        self.test_mode_enabled : bool = False
        self.listening : bool = False
        self.received_message_handler = received_message_handler
        self.medium.join(self)

    @override
    def check_connection(self) -> bool:
        return True

    @override
    def enable_test_mode(self) -> bool:
        self.test_mode_enabled = True

    @override
    def enable_listening(self) -> bool:
        """Enables listening. Can only do this in TEST mode."""

        if not self.test_mode_enabled:
            raise RuntimeError("TEST mode should be enabled before trying to listen.")
        self.listening = True

    @override
    def send_message(self, payload: bytes):
        """Sends a message over the medium. This action disables listening, and can only be done in TEST mode."""
        self.listening = False
        # Broadcast the message in a separate thread, so that it doesn't block the main one
        # It doesn't block either when sending over LoRa.
        messenger_thread = threading.Thread(
            target=self.medium.broadcast,
            args=(self, payload,)
        )
        messenger_thread.start()

    @override
    def handle_incoming_message_line(self, line):
        raise NotImplementedError(
            "This method is skipped in the mocked version." \
            "Messagess are directly given to the callback function as ReceivedMessage instances."
        )

    @override
    def _write_command_and_check_response(self, command, expected_response) -> bool:
        raise NotImplementedError()

    @override
    def is_listening(self) -> bool:
        return self.listening

if __name__ == '__main__':
    lora_controller = MockLoRaKitController()
    lora_controller.check_connection()
    lora_controller.enable_test_mode()
    lora_controller.enable_listening()

    lora_controller2 = MockLoRaKitController()
    lora_controller2.check_connection()
    lora_controller2.enable_test_mode()
    lora_controller2.send_message(b"HELLO WORLD")
    print("Did some work before other received my message")


    







from .received_message import ReceivedMessage

class ReceivedMessageDataParser:
    """Handles incoming message data lines. A message consists of 2 lines:

        +TEST: LEN:<length>, RSSI:<rssi>, SNR:<snr>

        +TEST: RX "<message>
    
        The lines always arrive in that order. So we make use of that.
        First we expect message metadata to arrive, then the payload.
        When the message is complete, it is returned as an instance of ReceivedMessage.
    """
    
    def __init__(self):
        self.message_getting_processed = None

    def process_message_line(self, message_line) -> None | ReceivedMessage:
        if self.message_getting_processed is None:
            # No message is already getting processed. So this line must be message metadata.
            metadata = self._parse_metadata(message_line)
            if not metadata:
                return
            # A new message is getting processed. Wait for the payload before returning it.
            self.message_getting_processed = ReceivedMessage(metadata)
            return
        else:
            # A message with metadata was already processed. So this must be the payload.
            payload = self._parse_payload(message_line)
            if not payload:
                return
            self.message_getting_processed.set_payload(payload)
            # Return the finished message, and reset the current message being handled.
            finished_message = self.message_getting_processed
            self.message_getting_processed = None
            return finished_message
            
    
    def _parse_metadata(self, metadata_line):
        """Handles a message line like:
        
        +TEST: LEN:<length>, RSSI:<rssi>, SNR:<snr>

        and extracts the metadata from it.
        """
        try: 
            metadata_line = metadata_line[len("+TEST: "):]  # Remove the "+TEST: " prefix for easier processing
            parts = metadata_line.split(",")
            metadata = {}
            for part in parts:
                key, value = part.split(":")  # Split each part into key and value
                metadata[key.strip()] = int(value.strip())  # Store in metadata dictionary
            return metadata
        except Exception:
            print(f"⚠️ WARNING: tried to parse a serial line as message metadata, but got: {metadata_line}")
            return None


    def _parse_payload(self, payload_line):
        """Handles a message line like:
        
        +TEST: RX "<message>"

        and extracts the payload from it.
        """
    
        splitted = payload_line.split("\"")
        if len(splitted) < 2:
            print(f"⚠️ WARNING: tried to parse a serial line as message payload, but got: {payload_line}")
            return None
        else:
            return splitted[1]


import pickle

from .received_message_data_parser import ReceivedMessage

class Response:
    """A response is a class that represents a message response to a previously sent message.
    A Response instance is sent over the medium, encoded as bytes."""

    def __init__(self, response_for: bytes | ReceivedMessage, response_contents: bytes):
        assert isinstance(response_for, bytes) or isinstance(response_for, ReceivedMessage), "response_for should be either bytes or a ReceivedMessage instance"
        if isinstance(response_for, ReceivedMessage):
            self.response_for = response_for.get_payload()
        else: 
            self.response_for = response_for
        self.response_contents = response_contents

    @staticmethod
    def from_bytes(bytes_response):
        """Tries to convert a response from raw bytes. If this didn't work None is retured."""
        try: 
            dict_instance: dict = pickle.loads(bytes_response)
            # Check if the received dict is convertible to a response
            if isinstance(dict_instance, dict):
                response_for_field = dict_instance.get('response_for')
                response_contents_field = dict_instance.get('response_contents')
                if len(dict_instance) == 2 and response_for_field is not None and response_contents_field is not None:
                    return Response(
                        response_for=response_for_field,
                        response_contents=response_contents_field
                    )
                else:
                    print(f"⚠️ WARNING: A response was received, but it could not be interpreted. Got: {dict_instance}")
                    return None
            else:
                return None
        except pickle.UnpicklingError: # Can be thrown by the pickle.loads().
            # So if the load didn't work, the message is probably not a response.
            return None

    def is_response_for(self, earlier_message: bytes) -> bool:
        """Check if this response was a response for a certain earlier message."""
        return self.response_for == earlier_message

    def as_bytes(self):
        response_dict = {'response_for': self.response_for, 'response_contents': self.response_contents}
        return pickle.dumps(response_dict)
    
    def get_original_message(self) -> bytes:
        """Get the message that this response responded to."""
        return self.response_for
    
    def get_contents(self) -> bytes:
        return self.response_contents
    
    def __repr__(self):
        return f"Response({self.response_for},{self.response_contents})"

if __name__ == '__main__':
    # Test if a response can be converted into bytes and back.
    message_payload = b'HELLO WORLD'
    response_payload = b'HELLO'

    response = Response(message_payload, response_payload)
    bytes_response = response.as_bytes()

    response2 = Response.from_bytes(bytes_response)

    assert response2.is_response_for(message_payload)
    print("Success")

    # Test if a Response can be created from a ReceivedMessage.
    message = ReceivedMessage({}, message_payload.hex())
    response = Response(message, response_payload)
    assert response.is_response_for(message.get_payload())
    print("Success")

    assert Response.from_bytes(message_payload) is None
    print("Success")

